import pdb
import numpy as np
import numpy.linalg as la
from util import hfs
import util

# converts f_occ array of particle numbers by block to orbital occupation 
# numbers, filling the first orbitals in each block
def expandNocc(nocc):
    occnums = np.zeros(hfs.Nbasis, dtype = int)
    ibn = 0 # Index of the current block in 'occnums'
    for ib in range(hfs.Nblock):
        Db = hfs.b2D[ib]
        occnumsb = np.zeros(Db, dtype = int)
        occnumsb[np.arange(nocc[ib])] = 1
        occnums[ibn:ibn + Db] = occnumsb
        ibn += Db
    return occnums

def calcHspx(rho):
  Hsp = util.calcT(rho) + util.calcV(rho)
  if hfs.lambdaq != 0.0:
    Hsp = Hsp + hfs.lambdaq*hfs.Qop
  return Hsp


#####
##### Routines to handle a single update step of the gradient minimization.
#####

def updateU(u, occnums, Horb):
    u_new = u*0.0
    u_z   = u*0.0
    idx = 0
    for ib in range(hfs.Nblock):
      Horb_b = util.unpackBlock(Horb,ib)
      u_b = util.unpackBlock(u,ib)
      M1 = Horb_b
      M2 = M1*1.0
      Db = np.shape(M1)[0]
      for i1 in range(Db):
        for i2 in range(i1+1,Db):
          M2[i1,i2] = M2[i1,i2]*hfs.eta
      evals,evecs = la.eigh(M2,'U')
      idx1 = idx+Db
      hfs.evalsL[idx:idx1] = evals
      idx = idx1 
      evecsT = np.transpose(evecs)
      u_c = np.dot(evecsT,u_b)
      util.packBlock(u_new,u_c,ib)
    return u_new

def projectZ3(Horb,u, focc):
  Horb2 = Horb.copy()
  Qorb = util.calcOrbOp(hfs.Qop, u)
  Qph = util.Op_ph(Qorb,focc)
  QDotQ = np.dot(Qph,Qph)               # full
  Hph = util.Op_ph(Horb,focc)    
  HDotQ = np.dot(Hph,Qph)
  mults = HDotQ/QDotQ
  Horb2 += -mults * Qorb
  return Horb2

def projectZ4(Horb,u, focc):
  Horb2 = Horb.copy()
  Qorb = util.calcOrbOp(hfs.Qop, u)
  Qph = Op_ph2(Qorb,focc)
  QDotQ = np.dot(Qorb,Qph)               # full
  HDotQ = np.dot(Horb,Qph)
#  pdb.set_trace()
  mults = HDotQ/QDotQ
  print 'mults2', mults
  Horb2 += -mults * Qorb
  return Horb2

def Op_ph2(Aorb,focc):    #   multiply (fa-fb) Aorb_ab
  Aph = Aorb*0.0
  for ib in range(hfs.Nblock):
    Ab = util.unpackBlock(Aorb, ib)
    Aphb = Ab*0.0
    Db = hfs.b2D[ib]
    for i1 in range(Db):
      for i2 in range(i1,Db):
         Aphb[i1,i2] = Ab[i1,i2]*(hfs.focc[hfs.b2idx01[ib]+i1] - hfs.focc[hfs.b2idx01[ib]+i2])
    util.packBlock(Aph,Aphb,ib)
  return Aph


def update_nocc(nocc):
  nocc_new = nocc*0
  return nocc_new

def sort(Evals,focc,nocc):
  nocc_new = nocc*0
  nn =0
  dec_n = [(E,i) for i, E in enumerate(Evals[:hfs.Nbasis_n])]
  dec_n.sort()
  for k in range(hfs.Nbasis_n):
    ib = hfs.o2b[dec_n[k][1]]
    nocc_new[ib] += 1      
    nn += 1
    if nn >= hfs.Nn:
      break
  np =0
  dec_p = [(E,i+hfs.Nbasis_n) for i, E in \
      enumerate(Evals[hfs.Nbasis_n:hfs.Nbasis])]
  dec_p.sort()
  for k in range(hfs.Nbasis-hfs.Nbasis_n):
    ix = dec_p[k][1]
    ib = hfs.o2b[ix]
    nocc_new[ib] += 1      
    np += 1
    if np >= hfs.Np:
      break
  newflag = False
  for k in range(hfs.Nblock):
    if nocc[k] != nocc_new[k]:
      hfs.nocc = nocc_new
      print 'new nocc'
      newflag = True
      break
  return newflag
  pdb.set_trace()
def resetQ(dQ,Qop,u,focc):
  Qorb = util.calcOrbOp(Qop, u)
  Qph = util.Op_ph(Qorb,focc)
  QDotQ = np.dot(Qph,Qph)		# full
  ddQ = dQ/QDotQ/2.      #  here we work with m>0 only
  Z = Qph*ddQ
  u_new = Qph*0.0
  for ib in range(hfs.Nblock):
    Zb = util.unpackBlock(Z,ib)
    D = np.shape(Zb)[0]
    uZbn = Zb/2.
    uZbd = -Zb/2.
    for ik in range(D):
      uZbn[ik,ik] += 1.0
      uZbd[ik,ik] += 1.0
    uZbd_inv = la.inv(uZbd)
    uZb = np.dot(uZbn,uZbd_inv)
    ub = util.unpackBlock(u,ib)
    ub_new = np.dot(uZb,ub)
    util.packBlock(u_new,ub_new,ib)
  return u_new

def entropy(focc):
  S= 0.0
  for ii in range(len(focc)):
    f = focc[ii]
    S += - f*np.log(f)-(1-f)*np.log(1-f)
  return 2*S                          # factor of 2 for +/- m  

def updatef(alpha_target,focc,Nbasis_n,Np,Nn):
   Nbasis = hfs.Nbasis
   alpha = f2a(focc)
   a_temp = hfs.eta_f*alpha_target + (1.0-hfs.eta_f)*alpha
   da0_min = -100.0
   da0_max = 100.0
   alpha_n = a_temp[:Nbasis_n]
   da0n = bisect(da0_min,da0_max,alpha_n,Nn/2)[0]
   alpha_p = a_temp[Nbasis_n:Nbasis]
   da0p = bisect(da0_min,da0_max,alpha_p,Np/2)[0]
   alpha[:Nbasis_n] = alpha_n + da0n
   alpha[Nbasis_n:Nbasis] = alpha_p + da0p
   focc1 = a2f(alpha)
   return focc1

def f2a(focc):
  alpha = focc*0.0
  for ii in range(hfs.Nbasis):
     alpha[ii] = np.log(1./focc[ii] - 1.0)
  return alpha
def a2f(alpha):
  focc = alpha*0.0
  for ii in range(hfs.Nbasis):
     focc[ii] = 1.0/(1.0 +np.e**(alpha[ii]))
  return focc

def bisect(x_min,x_max,alpha,N0):
# require function(x_max) > 0 and function(x_min) < 0
   for k in range(32):
     x_trial = (x_min+x_max)/2.
     if -(Nop_a(alpha,x_trial)-N0) > 0 :
        x_max = x_trial
     else:
        x_min = x_trial
   return x_min,x_max

def Nop_a(alpha,da0):
   N = 0.0
   for ii in range(len(alpha)):
     t1 = alpha[ii]+da0
     if t1> 25.:
       N += 0.0
     elif t1 < -25.:
       N += 1.0
     else:
       N += 1./(1.0 + np.e**(alpha[ii]+da0))
   return N
